;DCF77 Clock with extended display (14 digits, time with seconds and date)
;Controller: AT90S2313 or ATTiny2313 @ 4.194304 MHz external crystal
;
;Assemble with avra (http://avra.sourceforge.net/) only, this file contains
;directives specific for this assembler!
;
;(C) Arne Rossius, http://www.elektronik.de.vu
;
;DISPLAY:
;  - 14 LED displays, 7 segments + dp
;  - common cathode, segments connected through PNP transistors
;  - common cathode lines using 2x ULN2003 or compatible
;  - shift register for common lines, first output = left display
;  - shift register outouts are active high for ULN2003 inputs
;  - display order: HH:MM:SS YYYY-MM-DD
;
;PORTS:
;  PB0-7 = Segments (a..dp), active-low
;  PD2   = DCF77 input
;  PD5   = Data outout for shift register
;  PD6   = Clock output for shift register
;
;
;VERSION HISTORY:
;  v0.2  2007-01-10  Arne Rossius


;microcontroller definition file
.include "../include/2313def.inc"

;DCF77 receiver
.equ	DCFPort = PORTD
.equ	DCFPin = 2
.equ	DCFInvert = 1 ;inverted output (negative pulses), most receivers do this!

;LED display
.equ	SEGMENT = PORTB
.equ	COMMON = PORTD
.equ	DATA = 5
.equ	CLOCK = 6

;needed for lpm instruction
.def	lpmresult = R0

;registers for DCF77 telegram
.def	dcf1 = R3
.def	dcf2 = R4
.def	dcf3 = R5
.def	dcf4 = R6
.def	dcf5 = R7

;time/date registers
.def	subsec = R8
.def	second = R9
.def	minute = R10
.def	hour = R11
.def	day = R12
.def	month = R13
.def	year = R14
.def	wkday = R15 ;(weekday, 1=MO 7=SU)

;general registers
.def	temp = R16
.def	temp2 = R17

;DCF77-specific registers
.def	dcfbits = R18 ;received bits
.def	dcftime = R19 ;1/64-seconds since last edge
.def	flags1 = R20
.def	flags2 = R21

;additional registers for multiplexing
.def	display = R22 ;active display [0..13]

;bit definitions for the flags1 register
.equ	ErrBitCount = 0 ;last error: wrong number of bits
.equ	ErrParity = 1 ;last error: parity error
.equ	ErrBitLength = 2 ;last error: wrong bit length
.equ	ErrStartBit = 3 ;last error: start bit was not 1
.equ	TimeOK = 4 ;time is OK
.equ	DateOK = 5 ;date is OK (reset at midnight)
.equ	StopDCF = 6 ;stop DCF77 reception until a new telegram starts
.equ	DCFdisabled = 7 ;DCF77 receiver disabled (reset when a new hour begins)
.equ	ErrMask = 1<<ErrBitLength | 1<<ErrParity | 1<<ErrBitCount | 1<<ErrStartBit ;used to reset errors

;bit definitions for the flags2 register
.equ	LastState = DCFPin ;last state of DCF77 input, must be the same bit as on the port!

.cseg

;*******************************************************************************
;*
;*   MAIN PROGRAM
;*
;*******************************************************************************

;---------------------------------------
; Interrupt vectors
;---------------------------------------

;Reset
.org 0x000
	rjmp	start
	
;Timer 0 overflow
.org OVF0addr
	rjmp	ovf0int

;---------------------------------------
; Initialization
;---------------------------------------
	
start:
	;stack pointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;I/O ports
	ldi	temp, 0xFF
	out	DDRB, temp
	out	PORTB, temp
	ldi	temp, 0x60
	out	DDRD, temp
	ldi	temp, 0x1F
	out	PORTD, temp
	
	;timer
	ldi	temp, 1<<TOIE0 ;overflow int on
	out	TIMSK, temp
	ldi	temp, 0x04 ;Clk/256 = 64Hz int
	out	TCCR0, temp
	
	;registers
	ldi	dcfbits, 0
	ldi	dcftime, 0
	ldi	flags1, 1<<StopDCF
	ldi	flags2, 0
	ldi	display, 0
	
	;enable interrupts
	sei

;---------------------------------------
; Main loop
;   modifies: temp, temp2, display, common, ZL, ZH, lpmresult
;---------------------------------------

;multiplex all the time and do *all* the action while one of the displays is enabled
main_loop:
	;enable next display
	ldi	temp2, 0xFF
	out	SEGMENT, temp2
	sbi	COMMON, CLOCK
	cbi	COMMON, CLOCK
	out	SEGMENT, temp
	
	;encode the next display
	inc	display
	cbi	COMMON, DATA
	cpi	display, 14 ;return to display 0 after display 13
	brlo	PC+3
	ldi	display, 0
	sbi	COMMON, DATA
	
	;set data for display
	cpi	display, 2
	brlo	disp_hr
	cpi	display, 4
	brlo	disp_min
	cpi	display, 6
	brlo	disp_sec
	cpi	display, 8
	brlo	disp_20
	cpi	display, 10
	brlo	disp_yr
	cpi	display, 12
	brlo	disp_mon
	mov	ZL, day		;day (or DCF77 input)
	sbrs	flags1, DateOK
	ldi	ZL, 0xEE ;"--"
	sbrc	flags1, TimeOK
	rjmp	display_end
	ldi	ZL, 0xFD ;" _"
	sbrc	flags2, LastState
	ldi	ZL, 0xFE ;" -"
	rjmp	display_end
disp_mon:			;month
	mov	ZL, month
	sbrs	flags1, DateOK
	ldi	ZL, 0xEE ;"--"
	sbrs	flags1, TimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_yr:			;year
	mov	ZL, year
	sbrs	flags1, DateOK
	ldi	ZL, 0xEE ;"--"
	sbrs	flags1, TimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_20:			;century (20)
	ldi	ZL, 0x20
	sbrs	flags1, DateOK
	ldi	ZL, 0xEE ;"--"
	sbrs	flags1, TimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_sec:			;second (or last received bit)
	mov	ZL, second
	sbrc	flags1, TimeOK
	rjmp	display_end
	ldi	ZL, 0xFE ;" -"
	cpi	dcfbits, 0
	breq	display_end
	sbrc	flags1, StopDCF
	rjmp	display_end
	ldi	ZL, 0xF0 ;" 0"
	sbrc	dcf5, 7
	ldi	ZL, 0xF1 ;" 1"
	rjmp	display_end
disp_min:			;second (or last error)
	mov	ZL, minute
	sbrc	flags1, TimeOK
	rjmp	display_end
	ldi	ZL, 0xFE ;" -"
	sbrc	flags1, ErrBitCount
	ldi	ZL, 0xFA ;" C"
	sbrc	flags1, ErrBitLength
	ldi	ZL, 0xFB ;" L"
	sbrc	flags1, ErrParity
	ldi	ZL, 0xFC ;" P"
	sbrc	flags1, ErrStartBit
	ldi	ZL, 0xF5 ;" S"
	rjmp	display_end
disp_hr:			;hour (or # of bits received)
	mov	ZL, hour;
	sbrc	flags1, TimeOK
	rjmp	display_end
	mov	ZL, dcfbits
	rcall	bin2bcd
display_end:

	;suppress leading zero
	cpi	display, 0
	brne	PC+4
	cpi	ZL, 0x10
	brsh	PC+2
	ldi	ZL, 0xF0 ;15 = space
	
	;encode to 7-segment
	sbrs	display, 0 ;swap if 10's wanted
	swap	ZL
	andi	ZL, 0x0F
	clr	ZH
	ldi	temp, LOW(charset*2)
	add	ZL, temp
	ldi	temp, HIGH(charset*2)
	adc	ZH, temp
	lpm
	mov	temp, lpmresult
	
	;enable dp after hours, minutes, year and month
	sbrs	flags1, TimeOK
	rjmp	dp_end
	cpi	display, 5
	breq	dp_end
	cpi	display, 7
	breq	dp_end
	cpi	display, 13
	breq	dp_end
	sbrc	display, 0
	andi	temp, ~0x80
dp_end:
	
	;wait for a while
	ldi	ZH, 2
	ldi	ZL, 0
multiplex_wait:
	dec	ZL
	brne	multiplex_wait
	dec	ZH
	brne	multiplex_wait
	
	rjmp	main_loop
	
;---------------------------------------
; 7-Segment encoding table
;---------------------------------------

charset:
	;     0     1     2     3     4     5     6     7
	.db 0xC0, 0xF9, 0xA4, 0xB0, 0x99, 0x92, 0x82, 0xF8
	;     8     9     C     L     P     _     -    SPC
	.db 0x80, 0x90, 0xC6, 0xC7, 0x8C, 0xF7, 0xBF, 0xFF

;*******************************************************************************
;*
;*   DCF77 subroutines
;*
;*******************************************************************************

;---------------------------------------
; Timer 0 overflow int (64Hz)
;   modifies: dcftime, dcfbits, flags1, flags2, dcf1, dcf2, dcf3, dcf4, dcf5,
;             subsec, second, minute, hour, day, wkday, month, year
;---------------------------------------

ovf0int:
	;save registers
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
	
	;don't evaluate DCF77 signal when reception is disabled
	sbrc	flags1, DCFdisabled
	rjmp	dcf_end
	
	inc	dcftime
	
	;compare DCF77 input to last saved state
	in	temp, DCFPort-2 ;PINx
	eor	temp, flags2
	andi	temp, 1<<LastState
	.if ( DCFInvert )
		.message "DCF77 Signal: INVERTED"
		brne	dcf_end ;no change
	.else
		.message "DCF77 Signal: normal"
		breq	dcf_end ;no change
	.endif
	
	;take appropriate action for pos/neg edges
	sbrs	flags2, LastState
	rjmp	dcf_posedge
	
		; < negative edge ( 1 -> 0 ) >
	
	andi	flags2, ~(1<<LastState)
	
	;exit if an error was detected before
	sbrc	flags1, StopDCF
	rjmp	dcf_end
	 
	;increment bit number, exit if more than 59 bits were already received
	inc	dcfbits
	cpi	dcfbits, 60
	brlo	dcf_evalbit
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitCount | 1<<StopDCF
	rjmp	dcf_end
	
		; < positive edge ( 0 -> 1 ) >
dcf_posedge:
	ori	flags2, 1<<LastState
	
	;check if time was >1.2s (start of new telegram)
	cpi	dcftime, 77
	ldi	dcftime, 0
	brlo	dcf_end
	
	;end of telegram / new telegram
	andi	flags1, ~ErrMask & ~(1<<StopDCF)
	cpi	dcfbits, 59
	ldi	dcfbits, 0
	breq	dcf_eval
	ori	flags1, 1<<ErrBitCount
dcf_end:

;---------------------------------------
; increment time (fallback quartz clock)
;---------------------------------------

	;increment 1/64-seconds
	inc	subsec
	ldi	temp, 64
	cp	subsec, temp
	brlo	inctime_end
	
	;increment seconds
	clr	subsec
	mov	temp, second
	rcall	bcdinc
	mov	second, temp
	cpi	temp, 0x60
	brlo	inctime_end
	
	;increment minutes
	clr	second
	mov	temp, minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x04 ;disable DCF77 reception if nothing received for 4 minutes
	brne	inctime_nodisable
	sbrc	flags1, TimeOK
	ori	flags1, 1<<DCFdisabled
inctime_nodisable:
	cpi	temp, 0x60
	brlo	inctime_end

	;increment hours
	clr	minute
	andi	flags1, ~(1<<DCFdisabled) ;re-enable DCF77 reception
	mov	temp, hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x24
	brlo	inctime_end
	
	;new day: mark date as bad until the new one was received
	clr	hour
	andi	flags1, ~(1<<DateOK)
inctime_end:

	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti

;---------------------------------------
; DCF77 reception (continued)
;---------------------------------------

		; < negative edge - cont'd >
	
	;check bit length (optional)
dcf_evalbit:
	cpi	dcftime, 5 ;bit is 0: 5 <= dcftime <= 7
	brlo	dcf_errbit
	cpi	dcftime, 8
	brlo	dcf_clearbit
	cpi	dcftime, 11 ;bit is 1: 11 <= dcftime <= 13
	brlo	dcf_errbit
	cpi	dcftime, 14
	brlo	dcf_setbit
	
	;received bit is erroneous
dcf_errbit:
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitLength | 1<<StopDCF
	rjmp	dcf_end
	
	;save bit
dcf_clearbit:
	clc ;clear bit
	rjmp	dcf_savebit
dcf_setbit:
	sec ;set bit
dcf_savebit:
	ror	dcf5
	ror	dcf4
	ror	dcf3
	ror	dcf2
	ror	dcf1
	rjmp	dcf_end
	
;---------------------------------------
; evaluate DCF77 telegram
;---------------------------------------
	
dcf_eval:
	;check date parity
	ldi	temp2, 0
	mov	temp, dcf3
	andi	temp, 0xFE
	rcall	parity
	mov	temp, dcf4
	rcall	parity
	mov	temp, dcf5
	andi	temp, 0x7F
	rcall	parity
	mov	temp, dcf5
	lsl	temp ;Bit7 -> C
	rol	temp ;C -> Bit0
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check hour parity
	ldi	temp2, 0
	mov	temp, dcf2
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf3
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check minute parity
	ldi	temp2, 0
	mov	temp, dcf1
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf2
	andi	temp, 0x01
	rcall	parity
	mov	temp, dcf2
	lsr	temp
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror

	;save minutes
	lsr	dcf2
	ror	dcf1
	lsr	dcf1
	mov	minute, dcf1
	
	;save hours
	lsr	dcf2
	mov	hour, dcf2
	
	;save day
	lsl	dcf3 ;align day
	rol	dcf4 ;align month[5] weekday[3]
	rol	dcf5 ;align year
	lsr	dcf3
	lsr	dcf3
	mov	day, dcf3
	
	;save weekday
	mov	wkday, dcf4
	ldi	temp, 0x07
	and	wkday, temp
	
	;save month
	lsr	dcf4
	lsr	dcf4
	lsr	dcf4
	mov	month, dcf4
	
	;save year
	mov	year, dcf5
	
	;set time & date ok, disable DCF77 reception
	clr	subsec
	clr	second
	ori	flags1, 1<<TimeOK | 1<<DateOK | 1<<DCFdisabled
	rjmp	dcf_end
	
dcf_parityerror:
	;set parity error
	ori	flags1, 1<<ErrParity
	rjmp	dcf_end

;*******************************************************************************
;*
;*   GENERAL ARITHMETIC FUNCTIONS
;*
;*******************************************************************************

;---------------------------------------
; calculate parity and add up
;   in: temp
;   out: temp2 (add to existing value)
;---------------------------------------

parity:
	lsr	temp
	brcc	PC+2
	inc	temp2
	cpi	temp, 0
	brne	parity
	andi	temp2, 0x01
	ret
	
;---------------------------------------
; increment packed bcd number
;   in: temp
;   out: temp
;---------------------------------------

bcdinc:
	subi	temp, -7 ;increment and add 6
	brhc	PC+2 ;half-carry is inverted when using "subi rh, -k" instead of add!
	subi	temp, 6
	ret
	
;---------------------------------------
; convert binary to packed BCD
;   in: ZL
;   out: ZL
;   modifies: temp
;---------------------------------------

bin2bcd:
	ldi	temp, 0
bin2bcd_loop:
	subi	temp, -0x10
	subi	ZL, 10
	brcc	bin2bcd_loop
	subi	temp, 0x10
	subi	ZL, -10
	add	ZL, temp
	ret